@@ -19,11 +19,16 @@ from requests import get as _get
 from bs4 import BeautifulSoup as _BeautifulSoup
 from math import ceil as _ceil
 import user_agent
+from .errors import *
+
+""" ERRORS EXCEPTIONS """
 
 
 class Github:
     """ Pass a Github username to the object of this class to begin collect information!"""
 
+    current_path = _os.getcwd()  # class attribute.
+
     def __init__(self, username):
         self.username = username
         self.url = f"https://github.com/{self.username}"
@@ -32,9 +37,10 @@ class Github:
         }
 
         if not _get(f"https://github.com/{self.username}"):
-            exit("User is NOT existed!")
+            raise UsernameNotFound(f"{self.username} is NOT existed on Github.")
 
-        self.__resp = _get(f"https://api.github.com/users/{self.username}").json()
+        self.__resp = _get(f"https://api.github.com/users/{self.username}",
+                           headers={"User-Agent": user_agent.generate_user_agent()}).json()
 
         try:
             self.userId = self.__resp["id"]
@@ -42,19 +48,34 @@ class Github:
             self.userId = None
 
         self.__headers = {"Host": "github.com",
-                        "User-Agent": user_agent.generate_user_agent(),
-                        "Accept-Language": "en-US,en;q=0.5",
-                        "Accept-Encoding": "gzip, deflate, br",
-                        "Referer": "https://github.com",
-                        "Connection": "keep-alive",
-                        "Cookie": "user_session=kwO7LcPSqCVUguEf7xonR6ycKmuqm8XvI0qVco0FjE1ZKY1w;"
-
-                        }
+                          "User-Agent": user_agent.generate_user_agent(),
+                          "Accept-Language": "en-US,en;q=0.5",
+                          "Accept-Encoding": "gzip, deflate, br",
+                          "Referer": "https://github.com",
+                          "Connection": "keep-alive",
+                          "Cookie": "user_session=kwO7LcPSqCVUguEf7xonR6ycKmuqm8XvI0qVco0FjE1ZKY1w;"}
 
         try:
-            self.countryName = str(self.__resp["location"]).split(",")[1].strip()
-        except Exception as e:
-            self.countryName = None
+            countryName = str(self.__resp["location"]).split(",")[1].strip()
+            if len(countryName.split(",")) == 2:
+                self.countryName = countryName.split(",")[1]
+            elif len(countryName.split(",")) == 1:
+                self.countryName = countryName
+        except Exception:
+            try:
+                countryName = _BeautifulSoup(_get(f"https://github.com/{self.username}",
+                                                  headers={"User-Agent": user_agent.generate_user_agent()}).text,
+                                             "html.parser").find("li", attrs={"itemprop": "homeLocation"}).find("span",
+                                                                                                                attrs={
+                                                                                                                    "class": "p-label"}).text
+                if len(countryName.split(",")) == 2:
+                    self.countryName = countryName.split(",")[1]
+                elif len(countryName.split(",")) == 1:
+                    self.countryName = countryName
+                else:
+                    self.countryName = None
+            except (AttributeError, IndexError):
+                self.countryName = None
 
         if self.username.strip() == "":
             print("Github username must NOT be empty.")
@@ -69,7 +90,7 @@ class Github:
         #     self.followers = None
 
         try:
-            self.rep_os = self.__resp["public_rep_os"]
+            self.rep_os = self.__resp["public_repos"]
         except:
             self.rep_os = None
 
@@ -216,30 +237,24 @@ class Github:
             return None
         return countryinfo
 
-    current_path = _os.getcwd()
-
-    def downloadProfilePic(self, Path=current_path):
+    def downloadProfilePic(self, user=None, Path=current_path):
         """ Method for downloading specific  profile picture of Github user or Id
          You can specify the path for downloading the picture, default is current path.
          """
-        if not _os.path.exists(Path):
-            _os.mkdir(Path)
+        if user is None:  # download object profile picture.
+            user = self.username
+            picUrl = f"https://avatars.githubusercontent.com/u/{self.__resp['id']}?v=4"  # The profile pic of the
+            # object or instance.
+        else:  # download the profile picture of another user
+            picUrl = _BeautifulSoup(_get(f"https://github.com/{user}", headers=self.__headers).text,
+                                    "html.parser").find("img", attrs={
+                "class": "avatar avatar-user width-full border color-bg-primary"})["src"]
 
-        _os.chdir(Path)
+        self.__createPathAndEnterIn(Path)
+        picName = user + ".jpeg"
 
-        picUrl = f"https://avatars.githubusercontent.com/u/{self.__resp['id']}?v=4"  # The profile pic of the
-        # object or instance.
-        picName = self.username + ".jpeg"
-
-        # if not _os.path.isfile(_os.path.join(_os.path.abspath(Path), picName)):
         if not _pathlib.Path(picName).exists():
             _urlretrieve(picUrl, picName)
-            """
-            print(_pathlib.Path(picName))
-            print(_pathlib.Path(picName).exists())
-            """
-        else:
-            pass
         _os.chdir(self.CURRENT_PATH)
 
         return _os.path.abspath(_os.path.join(Path, picName))
@@ -256,7 +271,7 @@ class Github:
     """
 
     def getProfileById(self, UserId: str) -> str:
-        """ Method that _gets the user link profile by his Github id """
+        """ Method that gets the user link profile by his Github id """
         api = f"https://api.github.com/user/{UserId}"  # API for return user info by ID.
         profileLink = _get(api).json()["html_url"]
         return profileLink
@@ -277,11 +292,10 @@ class Github:
 
     def getFollowersProfiles(self) -> list:
         """ Simple method for fetching followers urls """
-        followersProfiles = []
-        followersProfilesResponse = _get(self.__resp["followers_url"]).json()
-        for item in followersProfilesResponse:
-            followersProfiles.append(item["html_url"])
-        return followersProfiles
+        return [f"https://github.com/{followerUsername}" for followerUsername in self.getFollowersUsernames()]
+
+    def getFollowingProfiles(self) -> list:
+        return [f"https://github.com/{followingUsername}" for followingUsername in self.getFollowingUsernames()]
 
     def getFollowingImgs(self) -> list:
         """ _get All Following Pictures """
@@ -312,32 +326,30 @@ class Github:
 
         self.__createPathAndEnterIn(Path)
 
-        followersProfilesPicsList = self._getFollowersImgs()
+        followersProfilesPicsList = self.getFollowersImgs()
+        followersUsernames = self.getFollowersUsernames()
 
         """ Downloading Profiles Pictures """
-        for imageLink in followersProfilesPicsList:
-            imgName = (str(imageLink).split("/")[-1])[:-4] + ".jpeg"
+        for imageLink, followerUsername in zip(followersProfilesPicsList, followersUsernames):
+            # imgName = (str(imageLink).split("/")[-1])[:-4] + ".jpeg"
+            imgName = followerUsername + ".jpeg"  # profile picture will be named with it's owner.
             if not _pathlib.Path(imgName).exists():
                 _urlretrieve(imageLink, imgName)
-            else:
-                # print(f"image {imgName} is already exists")
-                pass
         _os.chdir(self.CURRENT_PATH)
 
     def downloadFollowingPics(self, Path=current_path):
         """ Method for downloading following profile phot_os  """
         self.__createPathAndEnterIn(Path)
 
-        followingsProfilesPicsList = self._getFollowingProfilesPics()
-
+        followingsProfilesPicsList = self.getFollowingImgs()
+        followingUsernames = self.getFollowingUsernames()
         """ Downloading Profiles Pictures """
-        for imageLink in followingsProfilesPicsList:
-            imgName = (str(imageLink).split("/")[-1])[:-4] + ".jpeg"
+        for imageLink, followingUsername in zip(followingsProfilesPicsList, followingUsernames):
+            # imgName = (str(imageLink).split("/")[-1])[:-4] + ".jpeg"
+            imgName = followingUsername + ".jpeg"  # profile picture will be named with it's owner.
             if not _pathlib.Path(imgName).exists():
                 _urlretrieve(imageLink, imgName)
-            else:
-                # print(f"image {imgName} is already exists")
-                pass
+
         _os.chdir(self.CURRENT_PATH)
 
     def __path__(self):
@@ -367,7 +379,7 @@ class Github:
 
     def userEmail(self, user=None):
         """
-        :return user the email of a specific user if provided
+        :return email of a specific user if provided
         otherwise returns the email of object
         """
         if user is None:
@@ -375,13 +387,42 @@ class Github:
         soup = _BeautifulSoup(_get(f"https://github.com/{user}", headers=self.__headers).text, "html.parser")
         try:
             userEmail = soup.find('a', {"class": "u-email Link--primary"}).text
-        except AttributeError:
+        except AttributeError:  # email is NOT provided.
             userEmail = None
 
         return userEmail
 
+    def allEmails(self, user=None) -> list:
+        """
+        :return all available emails of a specific user if provided
+        otherwise returns the email of object's user.
+        """
+        if user is None:
+            user = self.username
+        emails = [emailTag["href"][7:] for emailTag in _BeautifulSoup(
+            _get(f"https://github.com/{user}", headers={"User-Agent": user_agent.generate_user_agent()}).text,
+            "html.parser").select("a[href ^= 'mailto:']")]
+        return emails
+
+    def extLinks(self, user=None) -> list:
+        """ :return the external links of Github user
+        if user argument is None method will return t-
+        he external links of object's user.
+        """
+        if user is None:
+            user = self.username
+
+        extLinks = [aTag["href"] for aTag in
+                    _BeautifulSoup(_get(f"https://github.com/{user}",
+                                        headers={"User-Agent": user_agent.generate_user_agent()}).text,
+                                   "html.parser").find(
+                        "div", attrs={"class": "js-profile-editable-area d-flex flex-column d-md-block"}).find_all("a",
+                                                                                                                   attrs={
+                                                                                                                       "class": "Link--primary"})]
+        return extLinks
+
     def getFollowersUsernames(self) -> list:
-        """get all followers usernames"""
+        """ get all followers usernames """
 
         followersUsernames = []
 
@@ -401,7 +442,10 @@ class Github:
 
         return followersUsernames
 
-    def getFollowersEmails(self) -> list:
+    # def getFollowersUsernames(self):
+    #     _Thread(target=self._getFollowersUsernames).start()
+
+    def getFollowersEmails(self, output=None) -> list:
         """get all followers emails"""
 
         followersUsernames = self.getFollowersUsernames()
@@ -409,6 +453,11 @@ class Github:
         for followerUsername in followersUsernames:
             followersEmails.append({followerUsername: self.userEmail(followerUsername)})
 
+        # Saving emails inside an output file
+        if output is not None:
+            for followerEmail in followersEmails:
+                with open(output, "a") as outputFile:
+                    print(followerEmail, file=outputFile)
         return followersEmails
 
     def getFollowingUsernames(self) -> list:
@@ -432,16 +481,23 @@ class Github:
 
         return followingUsernames
 
-    def getFollowingEmails(self) -> list:
+    def getFollowingEmails(self, output=None) -> list:
         """get all following emails"""
         followingUsernames = self.getFollowingUsernames()
         followingEmails = []
         for followingUsername in followingUsernames:
             followingEmails.append({followingUsername: self.userEmail(followingUsername)})
 
+        # Saving emails inside an output file
+        if output is not None:
+            for followingEmail in followingEmails:
+                with open(output, "a") as outputFile:
+                    print(followingEmail, file=outputFile)
         return followingEmails
 
+
 """
+/* TESTING/DEBUGGING */
 ################################################## Begin User Info ##################################################
 
 """
